/**
 * @file MotionFile.cpp
 * @brief [Vf[^ǂݏ֐Q`
 * @author cherub
 */
#include "stdafx.h"
#include <vector>
#include <map>
#include <string>
#include "MotionFile.h"

/**
 * @namespace NMotionFileLocal MotionFile.cpp
 * @brief [Vt@C͗pʃTu֐Q
 * @author cherub
 */
namespace NMotionFileLocal
{
	/**
	 * @enum channel_order
	 * @brief ʒuC]`l[_[
	 */
	enum channel_order
	{
		ORDER_RX = 1, ORDER_RY = 2, ORDER_RZ = 3,
		ORDER_TX = 4, ORDER_TY = 5, ORDER_TZ = 6
	};

	// g[N؂öׂ̋؂蕶
	const char *s_delim = "\t \n";
	// St@Cë̕ւ̊i[
	char* load_file(const char *fpath);
	// 3xNgf[^̓ǂݏo
	inline bool parse_vector_data(D3DXVECTOR3 &v);
	// En<-->n ]sϊ
	D3DXMATRIX basis_transform(const D3DXMATRIX &m);
	// En<-->n xNgϊ
	D3DXVECTOR3 basis_transform(const D3DXVECTOR3 &v);
};
using namespace NMotionFileLocal;


/**
 * @def GET_TOKEN(token, eaction)
 * @brief g[N擾}N
 * @param token [out] g[Ni[敶|C^
 * @param eaction [in] 擾s̏
 */
#define GET_TOKEN(token, eaction) { if (token = strtok(NULL, s_delim), token == NULL) eaction; }
/**
 * @def GET_COMP_TOKEN(token, compare, slen, eaction)
 * @brief g[N擾 & r}N
 * @param token [out] g[Ni[敶|C^
 * @param compare [in] r
 * @param slen [in] őr
 * @param eaction [in] 擾sorsv̏
 */
#define GET_COMP_TOKEN(token, compare, slen, eaction) { if (token = strtok(NULL, s_delim), token == NULL || strncmp(token, compare, slen) != 0) eaction; }

/**
 * @fn char* NMotionFileLocal::load_file(const char *fpath)
 * @brief St@Cë̕ւ̊i[
 * @param fpath [in] t@CpX
 * @return 擾擪ւ̃|C^i̓[U[̐ӔCj
 */
char* NMotionFileLocal::load_file(const char *fpath)
{
	FILE *fin = fopen(fpath, "r");
	if (fin == NULL)
		return NULL;
	fseek(fin, 0, SEEK_END);
	size_t fsize = ftell(fin);
	fseek(fin, 0, SEEK_SET);

	char *buf = new char[fsize];
	if (buf == NULL)
	{
		fclose(fin);
		return NULL;
	}

	fread(buf, sizeof(char), fsize, fin);
	fclose(fin);

	return buf;
}

/**
 * @fn D3DXMATRIX NMotionFileLocal::basis_transform(const D3DXMATRIX &m)
 * @brief En<-->n ]sϊ
 * @param m [in] ϊ]s
 * @return ϊ̉]s
 */
D3DXMATRIX NMotionFileLocal::basis_transform(const D3DXMATRIX &m)
{
	static D3DXMATRIX scale(1.0f, 0, 0, 0,  0, 1.0f, 0, 0,  0, 0, -1.0f, 0,  0, 0, 0, 1.0f);
	return scale * m * scale;
}

/**
 * @fn D3DXVECTOR3 NMotionFileLocal::basis_transform(const D3DXVECTOR3 &v)
 * @brief En<-->n xNgϊ
 * @param v [in] ϊxNg
 * @return ϊ̃xNg
 */
D3DXVECTOR3 NMotionFileLocal::basis_transform(const D3DXVECTOR3 &v)
{
	static D3DXMATRIX scale(1.0f, 0, 0, 0,  0, 1.0f, 0, 0,  0, 0, -1.0f, 0,  0, 0, 0, 1.0f);
	D3DXVECTOR3 vt;
	return *D3DXVec3TransformCoord(&vt, &v, &scale);
}

/**
 * @fn inline bool NMotionFileLocal::parse_vector_data(D3DXVECTOR3 &v)
 * @brief 3xNgf[^̓ǂݏo
 * @param v [out] 3DxNgf[^
 * @retval true ǂݏo
 * @retval false ǂݏos
 */
inline bool NMotionFileLocal::parse_vector_data(D3DXVECTOR3 &v)
{
	char *token;
	size_t i;
	for (i = 0;i < 3;++i)
	{
		GET_TOKEN(token, break);
		v[i] = static_cast<float>(atof(token));
	}
	return i == 3 ? true : false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @namespace NBvhFile MotionFile.cpp
 * @brief BVHt@C͗pTu֐Q
 * @author cherub
 */
namespace NBvhFile
{
	CJoint* parse_hierarchy(std::vector<int> &order);
	CJoint* parse_node(CJoint *parent, std::vector<int> &order);
	CJoint* parse_endnode(CJoint *parent);
	bool parse_channel_order(size_t channels, int &order);

	bool parse_motion(CMotionData *mot, const std::vector<int> &orders);
	bool parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q);
	bool parse_joint_rotation(int order, D3DXQUATERNION &q);

	bool write_hierarchy(FILE *fout, const CFigure *fig);
	bool write_sub_hierarchy(FILE *fout, const CJoint *node, size_t depth);
	bool write_motion(FILE *fout, const CFigure *fig, const CMotionData *mot);

	void get_euler_zxy(float &rz, float &rx, float &ry, const D3DXQUATERNION &q);
};

/**
 * @fn bool NBvhFile::parse_channel_order(size_t channels, int &order)
 * @brief `l̈ʒuC]I[_[ʎq̉
 * @param channels [in] `l
 * @param order [out] ʒuC]I[_[ʎq
 * @retval true ͐
 * @retval false ͎s
 */
bool NBvhFile::parse_channel_order(size_t channels, int &order)
{
	if (channels < 1 || channels > 6)
		return false;

	order = 0;

	char *token;
	for (size_t i = 0;i < channels;++i)
	{
		GET_TOKEN(token, return false);

		int norder = 0;
		if (strncmp(token, "Xposition", 9) == 0)
			norder = ORDER_TX;
		else if (strncmp(token, "Yposition", 9) == 0)
			norder = ORDER_TY;
		else if (strncmp(token, "Zposition", 9) == 0)
			norder = ORDER_TZ;
		else if (strncmp(token, "Xrotation", 9) == 0)
			norder = ORDER_RX;
		else if (strncmp(token, "Yrotation", 9) == 0)
			norder = ORDER_RY;
		else if (strncmp(token, "Zrotation", 9) == 0)
			norder = ORDER_RZ;
		else
			return false;

		order = order * 10 + norder;
	}

	return true;
}

/**
 * @fn CJoint* NBvhFile::parse_endnode(CJoint *parent)
 * @brief GhTCg
 * @param parent [in] e֐߃m[hւ̃|C^
 * @return ꂽ֐߃m[hւ̃|C^
 */
CJoint* NBvhFile::parse_endnode(CJoint *parent)
{
	char *token, *jname;
	D3DXVECTOR3 offset(0, 0, 0);

	GET_TOKEN(jname, return NULL);
	GET_COMP_TOKEN(token, "{", 1, return NULL);

	CJoint *node = new CJoint(offset, parent);
	node->SetName(jname);
	while (1)
	{
		GET_COMP_TOKEN(token, "OFFSET", 6, break);
		if (!parse_vector_data(offset))
			break;
		offset = basis_transform(offset);

		GET_COMP_TOKEN(token, "}", 1, break);

		node->SetOffset(offset);
		parent->AddChild(node);
		return node;
	}
	node->Destroy();
	return NULL;
}

/**
 * @fn CJoint* NBvhFile::parse_node(CJoint *parent, std::vector<int> &orders)
 * @brief ֐߃m[h
 * @param parent [in] e֐߃m[hւ̃|C^
 * @param orders [in/out] ]I[_[ʎq
 * @return ꂽ֐߃m[hւ̃|C^
 */
CJoint* NBvhFile::parse_node(CJoint *parent, std::vector<int> &orders)
{
	char *token, *jname;
	D3DXVECTOR3 offset(0, 0, 0);
	int order = 0;

	GET_TOKEN(jname, return NULL);
	GET_COMP_TOKEN(token, "{", 1, return NULL);

	CJoint *node = new CJoint(offset, parent);
	node->SetName(jname);
	while (token = strtok(NULL, s_delim), token != NULL)
	{
		if (strncmp(token, "OFFSET", 6) == 0)
		{
			if (!parse_vector_data(offset))
				break;
			offset = basis_transform(offset);

			GET_COMP_TOKEN(token, "CHANNELS", 8, break);
			GET_TOKEN(token, break);
			if (!parse_channel_order(static_cast<size_t>(atoi(token)), order))
				break;

			orders.push_back(order);
		}
		else if (strncmp(token, "JOINT", 5) == 0)
		{
			parse_node(node, orders);
		}
		else if (strncmp(token, "End", 3) == 0)
		{
			parse_endnode(node);
			orders.push_back(0);
		}
		else if (strncmp(token, "}", 1) == 0)
		{
			node->SetOffset(offset);
			if (parent != NULL)
				parent->AddChild(node);
			return node;
		}
		else
			break;
	}
	node->Destroy();

	return NULL;
}

/**
 * @fn CJoint* NBvhFile::parse_hierarchy(std::vector<int> &orders)
 * @brief XPgKw\̉
 * @param orders [in/out] ]I[_[ʎq
 * @return [gm[hւ̃|C^
 */
CJoint* NBvhFile::parse_hierarchy(std::vector<int> &orders)
{
	char *token;

	GET_COMP_TOKEN(token, "ROOT", 4, return NULL);
	return parse_node(NULL, orders);
}

/**
 * @fn bool NBvhFile::parse_joint_rotation(int order, D3DXQUATERNION &q)
 * @brief ֐߉]̉
 * @param order [in] ]I[_[ʎq
 * @param q [out] ]NH[^jI
 * @retval true ͐
 * @retval false ͎s
 */
bool NBvhFile::parse_joint_rotation(int order, D3DXQUATERNION &q)
{
	char *token;
	D3DXMATRIX m, mt;
	float value;

	D3DXMatrixIdentity(&m);
	for (int l = 100;l >= 1;l /= 10)
	{
		if (order >= l)
		{
			GET_TOKEN(token, return false);
			value = D3DXToRadian( static_cast<float>(atof(token)) );

			switch ((order / l)  % 10)
			{
			case ORDER_RX: D3DXMatrixRotationX(&mt, value); break;
			case ORDER_RY: D3DXMatrixRotationY(&mt, value); break;
			case ORDER_RZ: D3DXMatrixRotationZ(&mt, value); break;
			default: return false;
			}
			m = mt * m;
		}
	}
	m = basis_transform(m);

	D3DXQuaternionRotationMatrix(&q, &m);

	return true;
}

/**
 * @fn bool NBvhFile::parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q)
 * @brief [gʒuC]̉
 * @param order [in] ʒuI[_[ʎq
 * @param v [out] ʒuxNg
 * @param q [out] ]NH[^jI
 * @retval true ͐
 * @retval false ͎s
 */
bool NBvhFile::parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q)
{
	char *token;
	D3DXMATRIX m, mt;
	float value;

	v.x = v.y = v.z = 0;
	D3DXMatrixIdentity(&m);

	for (int l = 100000;l >= 1;l /= 10)
	{
		if (order >= l)
		{
			GET_TOKEN(token, return false);
			value = static_cast<float>(atof(token));

			D3DXMatrixIdentity(&mt);
			switch ((order / l) % 10)
			{
			case ORDER_TX: v.x = value; break;
			case ORDER_TY: v.y = value; break;
			case ORDER_TZ: v.z = value; break;
			case ORDER_RX: D3DXMatrixRotationX(&mt, D3DXToRadian(value)); break;
			case ORDER_RY: D3DXMatrixRotationY(&mt, D3DXToRadian(value)); break;
			case ORDER_RZ: D3DXMatrixRotationZ(&mt, D3DXToRadian(value)); break;
			default: return false;
			}
			m = mt * m;
		}
	}
	m = basis_transform(m);
	v = basis_transform(v);
	D3DXQuaternionRotationMatrix(&q, &m);
	return true;
}

/**
 * @fn bool NBvhFile::parse_motion(CMotionData *mot, const std::vector<int> &orders)
 * @brief BVH[Vf[^̉
 * @param mot [out] [Vf[^
 * @param orders [in] ʒuC]I[_[ʎq
 * @retval true ͐
 * @retval false ͎s
 */
bool NBvhFile::parse_motion(CMotionData *mot, const std::vector<int> &orders)
{
	char *token;
	GET_COMP_TOKEN(token, "Frames:", 7, return false);

	GET_TOKEN(token, return false);
	if (!mot->Initialize(static_cast<size_t>(atoi(token)), orders.size()))
		return false;

	GET_COMP_TOKEN(token, "Frame", 5, return false);
	GET_COMP_TOKEN(token, "Time:", 5, return false);
	GET_TOKEN(token, return false);

	D3DXVECTOR3 v;
	D3DXQUATERNION q;
	for (size_t f = 0;f < mot->NumFrames();++f)
	{
		if (!parse_root_coord(orders[0], v, q))
			return false;
		mot->SetPosition(f, v);
		mot->SetRotation(f, 0, q);

		for (size_t j = 1;j < mot->NumJoints();++j)
		{
			if (orders[j] == 0)
				D3DXQuaternionIdentity(&q);
			else if (!parse_joint_rotation(orders[j], q))
				return false;
			mot->SetRotation(f, j, q);
		}
	}
	return true;
}

/**
 * @fn bool NMotionFile::LoadBvhFile(const char *fpath, CFigure *fig, CMotionData *mot)
 * @brief BVHt@C̉
 * @param fpath [in] t@CpX
 * @param fig [out] XPgf[^
 * @param mot [out] [Vf[^
 * @retval true ͐
 * @retval false ͎s
 */
bool NMotionFile::LoadBvhFile(const char *fpath, CFigure *fig, CMotionData *mot)
{
	char *buf, *token;
	std::vector<int> orders;
	CJoint *root;
	bool bSuccess = false;

	if (buf = NMotionFileLocal::load_file(fpath), buf == NULL)
		return false;

	token = strtok(buf, s_delim);
	while (1)
	{
		if (strncmp(token, "HIERARCHY", 9) != 0)
			break;
		if (root = NBvhFile::parse_hierarchy(orders), root == NULL)
			break;
		fig->Initialize(root);

		GET_TOKEN(token, break);
		if (strncmp(token, "MOTION", 6) != 0)
			break;

		bSuccess = NBvhFile::parse_motion(mot, orders);
		break;
	}
	delete[] buf;

	return bSuccess;
}

/**
 * @fn bool NBvhFile::write_hierarchy(FILE *fout, const CFigure *fig)
 * @brief BVHXPgKw̃t@Co
 * @param fout [in] o͐BVHt@Cւ̃|C^
 * @param fig [in] XPgf[^
 * @retval true o͐
 * @retval false o͎s
 */
bool NBvhFile::write_hierarchy(FILE *fout, const CFigure *fig)
{
	const CJoint *root = fig->GetRoot();
	D3DXVECTOR3 offset = basis_transform(root->GetOffset());

	fprintf(fout, "HIERARCHY\n");
	fprintf(fout, "ROOT %s\n", root->GetName());
	fprintf(fout, "{\n");
	fprintf(fout, "  OFFSET %f %f %f\n", offset.x, offset.y, offset.z);
	fprintf(fout, "  CHANNELS 6 Xposition Yposition Zposition Zrotation Xrotation Yrotation\n");

	for (size_t i = 0;i < root->NumChildren();++i)
		write_sub_hierarchy(fout, root->GetChild(i), 1);
	fprintf(fout, "}\n");

	return true;
}

/**
 * @fn bool NBvhFile::write_sub_hierarchy(FILE *fout, const CJoint *node, size_t depth)
 * @brief BVHXPgKwo͊֐̃Tu֐
 * @param fout [in] o͐BVHt@Cւ̃|C^
 * @param node [in] ֐߃m[hւ̃|C^
 * @param depth [in] ċA[x
 * @retval true o͐
 * @retval false o͎s
 */
bool NBvhFile::write_sub_hierarchy(FILE *fout, const CJoint *node, size_t depth)
{
	D3DXVECTOR3 offset = basis_transform(node->GetOffset());

	for (size_t i = 0;i < depth * 2;++i) fputc(' ', fout);
	if (node->NumChildren() > 0)
		fprintf(fout, "JOINT %s\n", node->GetName());
	else
		fprintf(fout, "End %s\n", node->GetName());
	for (size_t i = 0;i < depth * 2;++i) fputc(' ', fout);
	fprintf(fout, "{\n");

	for (size_t i = 0;i < depth * 2 + 2;++i) fputc(' ', fout);
	fprintf(fout, "OFFSET %f %f %f\n", offset.x, offset.y, offset.z);
	if (node->NumChildren() > 0)
	{
		for (size_t i = 0;i < depth * 2 + 2;++i) fputc(' ', fout);
		fprintf(fout, "CHANNELS 3 Zrotation Xrotation Yrotation\n");
	}

	for (size_t i = 0;i < node->NumChildren();++i)
		write_sub_hierarchy(fout, node->GetChild(i), depth + 1);

	for (size_t i = 0;i < depth * 2;++i) fputc(' ', fout);
	fprintf(fout, "}\n");

	return true;
}
/**
 * @fn void NBvhFile::get_euler_zxy(float &ry, float &rx, float &rz, const D3DXQUATERNION &q)
 * @brief ]NH[^jICZXYI[_[̃IC[pɕ
 * @param rz [out] IC[p(Z)
 * @param rx [out] IC[p(X)
 * @param ry [out] IC[p(Y)
 * @param q [in] ]NH[^jI
 */
void NBvhFile::get_euler_zxy(float &rz, float &rx, float &ry, const D3DXQUATERNION &q)
{
	D3DXMATRIX m;
	D3DXMatrixRotationQuaternion(&m, &q);
	m = basis_transform(m);

	ry = atan2f(m._13, m._33);
	rx = -asinf(m._23);
	rz = atan2f(m._21, m._22);

	if (fabsf(cosf(rx)) < 1.0e-6f)
	{
		if (m._13 < 0.0f)
			ry = D3DX_PI + ry;
		else
			ry -= D3DX_PI;
		if (m._21 < 0.0f)
			rz = D3DX_PI + rz;
		else
			rz -= D3DX_PI;
	}

	rx = -D3DXToDegree(rx);
	ry = -D3DXToDegree(ry);
	rz = -D3DXToDegree(rz);
}

/**
 * @fn bool NBvhFile::write_motion(FILE *fout, const CFigure *fig, const CMotionData *mot)
 * @brief BVH[Ṽt@Co
 * @param fout [in] o͐BVHt@Cւ̃|C^
 * @param fig [in] XPgf[^
 * @param mot [in] [Vf[^
 * @retval true o͐
 * @retval false o͎s
 */
bool NBvhFile::write_motion(FILE *fout, const CFigure *fig, const CMotionData *mot)
{
	D3DXVECTOR3 v;
	D3DXQUATERNION q;
	D3DXMATRIX m;
	float rx, ry, rz;

	get_euler_zxy(rz, rx, ry, q);

	fprintf(fout, "MOTION\n");
	fprintf(fout, "Frames:  %d\n", mot->NumFrames());
	fprintf(fout, "Frame Time: 0.03333\n");

	for (size_t f = 0;f < mot->NumFrames();++f)
	{
		v = basis_transform(mot->GetPosition(f));
		fprintf(fout, "%f %f %f", v.x, v.y, v.z);

		for (size_t j = 0;j < mot->NumJoints();++j)
		{
			if (fig->IsEffector(j))
				continue;
			q = mot->GetRotation(f, j);
			get_euler_zxy(rz, rx, ry, q);

			fprintf(fout, " %f %f %f", rz, rx, ry);
		}
		fputc('\n', fout);
	}
	return true;
}

/**
 * @fn bool NMotionFile::StoreBvhFile(const char *fpath, const CFigure *fig, const CMotionData *mot)
 * @brief BVH[Vt@C̃Z[u
 * @param fpath [in] o͐BVHt@Cւ̃pX
 * @param fig [in] XPgf[^
 * @param mot [in] [Vf[^
 * @retval true Z[u
 * @retval false Z[us
 */
bool NMotionFile::StoreBvhFile(const char *fpath, const CFigure *fig, const CMotionData *mot)
{
	FILE *fout;
	if (fout = fopen(fpath, "w"), fout == NULL)
		return false;

	if (!NBvhFile::write_hierarchy(fout, fig))
		return false;
	if (!NBvhFile::write_motion(fout, fig, mot))
		return false;

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @namespace NAsfAmcFile MotionFile.cpp
 * @brief ASF-AMCt@C͗pTu֐Q
 * @author cherub
 */
namespace NAsfAmcFile
{
	CJoint* parse_asf(FILE *fin, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes);
	CJoint* parse_hierarchy(FILE *fin, std::map<std::string, CJoint*> &joints);
	CJoint* parse_asf_root(FILE *fin, std::map<std::string, int> &order);
	bool parse_asf_header(FILE *fin);
	bool parse_asf_bone(FILE *fin, std::map<std::string, CJoint*> &joints, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes);
	bool parse_root_order(int &order);
	bool parse_bone_order(int &order);
	bool parse_axis(D3DXMATRIX &m, const D3DXVECTOR3 &axis);
	D3DXMATRIX calc_axis_transform(CJoint *node, const std::map<std::string, D3DXMATRIX> &axes);

	bool parse_amc(FILE *fin, CMotionData *mot, CFigure *fig, const std::map<std::string, int> &orders, const std::map<std::string, D3DXMATRIX> &axes);
	bool parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q);
	bool parse_joint_rotation(int order, D3DXQUATERNION &q);
	size_t parse_frame_count(FILE *fin);

	void change_skeletal_structure(CJoint *node, D3DXVECTOR3 &v, std::map<std::string, D3DXMATRIX> &axes);
};

/**
 * @fn bool NAsfAmcFile::parse_joint_rotation(int order, D3DXQUATERNION &q)
 * @brief ֐߉]̉
 * @param order [in] ]I[_[ʎq
 * @param q [out] ]NH[^jI
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_joint_rotation(int order, D3DXQUATERNION &q)
{
	char *token;
	D3DXMATRIX m, mt;
	float value;

	D3DXMatrixIdentity(&m);
	for (int l = 100;l >= 1;l /= 10)
	{
		if (order >= l)
		{
			GET_TOKEN(token, return false);
			value = D3DXToRadian( static_cast<float>(atof(token)) );

			switch ((order / l)  % 10)
			{
			case ORDER_RX: D3DXMatrixRotationX(&mt, value); break;
			case ORDER_RY: D3DXMatrixRotationY(&mt, value); break;
			case ORDER_RZ: D3DXMatrixRotationZ(&mt, value); break;
			default: return false;
			}
			m *= mt;
		}
	}
	m = basis_transform(m);
	D3DXQuaternionRotationMatrix(&q, &m);

	return true;
}

/**
 * @fn bool NAsfAmcFile::parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q)
 * @brief [gʒuC]̉
 * @param order [in] ]I[_[ʎq
 * @param v [out] ʒuxNg
 * @param q [out] ]NH[^jI
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_root_coord(int order, D3DXVECTOR3 &v, D3DXQUATERNION &q)
{
	char *token;
	D3DXMATRIX m, mt;
	float value;

	v.x = v.y = v.z = 0;
	D3DXMatrixIdentity(&m);

	for (int l = 100000;l >= 1;l /= 10)
	{
		if (order >= l)
		{
			GET_TOKEN(token, return false);
			value = static_cast<float>(atof(token));

			D3DXMatrixIdentity(&mt);
			switch ((order / l) % 10)
			{
			case ORDER_TX: v.x = value; break;
			case ORDER_TY: v.y = value; break;
			case ORDER_TZ: v.z = value; break;
			case ORDER_RX: D3DXMatrixRotationX(&mt, D3DXToRadian(value)); break;
			case ORDER_RY: D3DXMatrixRotationY(&mt, D3DXToRadian(value)); break;
			case ORDER_RZ: D3DXMatrixRotationZ(&mt, D3DXToRadian(value)); break;
			default: return false;
			}
			m *= mt;
		}
	}
	v = basis_transform(v);
	m = basis_transform(m);
	D3DXQuaternionRotationMatrix(&q, &m);

	return true;
}

/**
 * @fn bool NAsfAmcFile::parse_asf_header(FILE *fin)
 * @brief ASFt@Cwb_̉
 * @param fin [in] ASFt@Cւ̃|C^
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_asf_header(FILE *fin)
{
	char buf[1024], *token;

	//version
	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':');
	if (token = strtok(buf, s_delim), token == NULL)
		return false;
	if (strncmp(token, ":version", 8) != 0) 
		return false;

	// name
	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':');
	if (token = strtok(buf, s_delim), token == NULL)
		return false;
	if (strncmp(token, ":name", 5) != 0)
		return false;

	// units
	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':');
	if (token = strtok(buf, s_delim), token == NULL)
		return false;
	if (strncmp(token, ":units", 6) != 0)
		return false;

	// documentation
	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':');
	if (token = strtok(buf, s_delim), token == NULL)
		return false;
	if (strncmp(token, ":documentation", 15) != 0)
		return false;

	return true;
}

/**
 * @fn bool NAsfAmcFile::parse_root_order(int &order)
 * @brief [g̈ʒuC]I[_[ʎq̉
 * @param order [out] ʒuC]I[_[ʎq
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_root_order(int &order)
{
	order = 0;

	char *token;
	while (token = strtok(NULL, s_delim), token != NULL)
	{
		int norder = 0;
		if (strncmp(token, "TX", 2) == 0)
			norder = ORDER_TX;
		else if (strncmp(token, "TY", 2) == 0)
			norder = ORDER_TY;
		else if (strncmp(token, "TZ", 2) == 0)
			norder = ORDER_TZ;
		else if (strncmp(token, "RX", 2) == 0)
			norder = ORDER_RX;
		else if (strncmp(token, "RY", 2) == 0)
			norder = ORDER_RY;
		else if (strncmp(token, "RZ", 2) == 0)
			norder = ORDER_RZ;
		else
			return false;

		order = order * 10 + norder;
	}

	return true;
}

/**
 * @fn bool NAsfAmcFile::parse_bone_order(int &order)
 * @brief ֐߉]I[_[ʎq̉
 * @param order [out] ]I[_[ʎq
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_bone_order(int &order)
{
	order = 0;

	char *token;
	while (token = strtok(NULL, s_delim), token != NULL)
	{
		int norder = 0;
		if (strncmp(token, "rx", 2) == 0)
			norder = ORDER_RX;
		else if (strncmp(token, "ry", 2) == 0)
			norder = ORDER_RY;
		else if (strncmp(token, "rz", 2) == 0)
			norder = ORDER_RZ;
		else
			return false;

		order = order * 10 + norder;
	}

	return true;
}

/**
 * @fn CJoint* NAsfAmcFile::parse_asf_root(FILE *fin, std::map<std::string, int> &orders)
 * @brief ASF[gm[h̉
 * @param fin [in] ASFt@Cւ̃|C^
 * @param orders [in/out] ֐ߖ-]I[_[̑Ή}bv
 * @return ꂽ[gm[hւ̃|C^
 */
CJoint* NAsfAmcFile::parse_asf_root(FILE *fin, std::map<std::string, int> &orders)
{
	char buf[1024], *token;
	D3DXVECTOR3 offset, orientation;

	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':')
	{
		if (token = strtok(buf, s_delim), token == NULL)
			return NULL;

		if (strncmp(token, "order", 5) == 0)
		{
			int order;
			parse_root_order(order);
			orders["root"] = order;
		}
		else if (strncmp(token, "axis", 4) == 0)
		{
			// TODO: implement (?)
		}
		else if (strncmp(token, "position", 8) == 0)
		{
			if (!parse_vector_data(offset))
				return false;
			offset = basis_transform(offset);
		}
		else if (strncmp(token, "orientation", 11) == 0)
		{
			if (!parse_vector_data(orientation))
				return false;
			orientation = basis_transform(orientation);
		}
		else
			return NULL;
	}

	if (token = strtok(buf, s_delim), token == NULL || strncmp(token, ":bonedata", 9) != 0)
		return NULL;
	CJoint *root = new CJoint(offset, NULL);
	root->SetName("root");
	return root;
}

/**
 * @fn bool NAsfAmcFile::parse_axis(D3DXMATRIX &m, const D3DXVECTOR3 &axis)
 * @brief axisI[_[̉͂axiss̎擾
 * @param m [out] axiss
 * @param axis [in] axis
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_axis(D3DXMATRIX &m, const D3DXVECTOR3 &axis)
{
	char *token;
	GET_TOKEN(token, false);

	D3DXMATRIX mt;
	D3DXMatrixIdentity(&m);
	size_t cnt;
	for (cnt = 0;cnt < 3;++cnt)
	{
		if (token[cnt] == 'X')
			D3DXMatrixRotationX(&mt, D3DXToRadian(axis[cnt]));
		else if (token[cnt] == 'Y')
			D3DXMatrixRotationY(&mt, D3DXToRadian(axis[cnt]));
		else if (token[cnt] == 'Z')
			D3DXMatrixRotationZ(&mt, D3DXToRadian(axis[cnt]));
		else
			break;

		m *= mt;
	}
	m = basis_transform(m);
	return cnt == 3 ? true : false;
}

/**
 * @fn bool NAsfAmcFile::parse_asf_bone(FILE *fin, std::map<std::string, CJoint*> &joints, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes)
 * @brief ASF{[̉
 * @param fin [in] ASFt@Cւ̃|C^
 * @param joints [in/out] ֐ߖ-֐߃m[h|C^̑Ή}bv
 * @param orders [in/out] ֐ߖ-֐߉]I[_[̑Ή}bv
 * @param axes [in/out] ֐ߖ-֐axiss̑Ή}bv
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_asf_bone(FILE *fin, std::map<std::string, CJoint*> &joints, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes)
{
	char buf[1024], *token, jname[256];
	CJoint *node;
	D3DXVECTOR3 dir, axis, v;
	D3DXMATRIX maxis, m;
	float len;
	int order = 0;

	while (fgets(buf, 1024, fin) != NULL)
	{
		token = strtok(buf, s_delim);
		if (strncmp(token, "id", 2) == 0)
		{
			GET_TOKEN(token, break);
		}
		else if (strncmp(token, "name", 4) == 0)
		{
			GET_TOKEN(token, break);
			strcpy(jname, token);
		}
		else if (strncmp(token, "direction", 9) == 0)
		{
			if (!parse_vector_data(dir))
				break;
			dir = basis_transform(dir);
		}
		else if (strncmp(token, "length", 6) == 0)
		{
			GET_TOKEN(token, break);
			len = static_cast<float>(atof(token));
		}
		else if (strncmp(token, "axis", 4) == 0)
		{
			if (!parse_vector_data(axis) || !parse_axis(maxis, axis))
				break;
			axes[jname] = maxis;
		}
		else if (strncmp(token, "dof", 3) == 0)
		{
			if (!parse_bone_order(order))
				break;
		}
		else if (strncmp(token, "limits", 6) == 0)
		{
			if (order >= 100)
				fgets(buf, 1024, fin);
			if (order >= 10)
				fgets(buf, 1024, fin);
		}
		else if (strncmp(token, "bodymass", 8) == 0)
		{
			// TODO: bodymass
		}
		else if (strncmp(token, "cofmass", 7) == 0)
		{
			// TODO: cofmass
		}
		else if (strncmp(token, "end", 3) == 0)
		{
			dir *= len;

			D3DXMatrixTranspose(&m, &maxis);
			D3DXVec3TransformCoord(&v, &dir, &m);

			node = new CJoint(v, NULL);
			node->SetName(jname);
			joints[jname] = node;
			orders[jname] = order;

			return true;
		}
		else
			break;
	}
	return false;
}

/**
 * @fn CJoint* NAsfAmcFile::parse_hierarchy(FILE *fin, std::map<std::string, CJoint*> &joints)
 * @brief ֐ߊKw\̉
 * @param fin [in] ASFt@Cւ̃|C^
 * @param joints [in/out] ֐ߖ-֐߃m[h|C^̑Ή}bv
 * @return [gm[hւ̃|C^
 */
CJoint* NAsfAmcFile::parse_hierarchy(FILE *fin, std::map<std::string, CJoint*> &joints)
{
	char buf[1024], *token;
	CJoint *root, *parent, *node;
	std::map<std::string, CJoint*>::iterator itj;

	if (fgets(buf, 1024, fin) == NULL)
		return NULL;
	if (token = strtok(buf, s_delim), token == NULL || strncmp(token, "begin", 5) != 0)
		return NULL;
	if (itj = joints.find("root"), itj == joints.end())
		return NULL;

	root = joints["root"];
	while (fgets(buf, 1024, fin) != NULL)
	{
		if (token = strtok(buf, s_delim), token == NULL)
			break;

		if (strncmp(token, "end", 3) == 0)
			return root;
		else
		{
			if (itj = joints.find(token), itj == joints.end())
				return NULL;
			parent = itj->second;
		}

		while (token = strtok(NULL, s_delim), token != NULL)
		{
			if (itj = joints.find(token), itj == joints.end())
				return NULL;
			node = itj->second;
			parent->AddChild(node);
		}
	}

	return NULL;
}

/**
 * @fn CJoint* NAsfAmcFile::parse_asf(FILE *fin, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes)
 * @brief ASFt@C̉
 * @param fin [in] ASFt@Cւ̃|C^
 * @param orders [out] ֐ߖ-֐߉]I[_[̑Ή}bv
 * @param axes [out] ֐ߖ-֐axiss̑Ή}bv
 * @return ꂽ[gm[hւ̃|C^
 */
CJoint* NAsfAmcFile::parse_asf(FILE *fin, std::map<std::string, int> &orders, std::map<std::string, D3DXMATRIX> &axes)
{
	char buf[1024], *token;
	CJoint *root;
	std::map<std::string, CJoint*> joints;
	D3DXMATRIX m;
	
	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':');
	token = strtok(buf, s_delim);
	if (strncmp(token, ":root", 15) != 0)
		return NULL;

	if (root = parse_asf_root(fin, orders), root == NULL)
		return NULL;
	joints["root"] = root;
	D3DXMatrixIdentity(&m);
	axes["root"] = m;

	while (fgets(buf, 1024, fin) != NULL && buf[0] != ':')
	{
		if (token = strtok(buf, s_delim), token == NULL ||  strncmp(token, "begin", 5) != 0)
			break;
		if (!parse_asf_bone(fin, joints, orders, axes))
			break;
	}

	if (token = strtok(buf, s_delim), token == NULL || strncmp(token, ":hierarchy", 10) != 0)
	{
		root->Destroy();
		for (std::map<std::string, CJoint*>::iterator it = joints.begin();it != joints.end();++it)
			it->second->Destroy();
		return NULL;
	}

	return parse_hierarchy(fin, joints);
}

/**
 * @fn size_t NAsfAmcFile::parse_frame_count(FILE *fin)
 * @brief AMCt[̉
 * @param fin [in] AMCt@Cւ̃|C^
 * @return t[
 */
size_t NAsfAmcFile::parse_frame_count(FILE *fin)
{
	char buf[1024], *token;
	size_t tmp, frames = 0;

	fseek(fin, 0L, SEEK_SET);
	while (fgets(buf, 1024, fin) != NULL)
	{
		if (token = strtok(buf, s_delim), token == NULL)
			return 0;

		tmp = static_cast<size_t>(atoi(token));
		if (tmp > frames)
			frames = tmp;
	}

	return frames;
}

/**
 * @fn D3DXMATRIX NAsfAmcFile::calc_axis_transform(CJoint *node, const std::map<std::string, D3DXMATRIX> &axes)
 * @brief axisɂgXtH[švZ
 * @param node [in] ֐߃m[hւ̃|C^
 * @param axes [in] ֐ߖ-֐axiss̑Ή}bv
 * @return axisgXtH[s
 */
D3DXMATRIX NAsfAmcFile::calc_axis_transform(CJoint *node, const std::map<std::string, D3DXMATRIX> &axes)
{
	const char *name1 = node->GetName();
	std::map<std::string, D3DXMATRIX>::const_iterator cit;

	cit = axes.find(name1);
	D3DXMATRIX m1 = cit->second;

	if (node->GetParent() != NULL)
	{
		const char *name2 = node->GetParent()->GetName();
		cit = axes.find(name2);
		D3DXMATRIX m2, mt = cit->second;
		D3DXMatrixTranspose(&m2, &mt);
		m1 = m1 * m2;
	}
	return m1;
}

/**
 * @fn bool NAsfAmcFile::parse_amc(FILE *fin, CMotionData *mot, CFigure *fig, const std::map<std::string, int> &orders, const std::map<std::string, D3DXMATRIX> &axes)
 * @brief AMCt@C̉
 * @param fin [in] AMCt@Cւ̃|C^
 * @param mot [out] [Vf[^
 * @param fig [in] XPgf[^
 * @param orders [in] ֐ߖ-֐߉]I[_[̑Ή}bv
 * @param axes [in/out] ֐ߖ-֐axiss̑Ή}bv
 * @retval true ͐
 * @retval false ͎s
 */
bool NAsfAmcFile::parse_amc(FILE *fin, CMotionData *mot, CFigure *fig, const std::map<std::string, int> &orders, const std::map<std::string, D3DXMATRIX> &axes)
{
	char buf[1024], *token;
	std::map<std::string, int>::const_iterator cit;
	size_t frames = parse_frame_count(fin), frm = 0;
	int jid;

	D3DXQUATERNION q;
	D3DXMATRIX m;
	D3DXVECTOR3 v;

	fseek(fin, 0L, SEEK_SET);

	if (!mot->Initialize(frames, fig->NumJoints()))
		return false;

	while (fgets(buf, 1024, fin) != NULL)
	{
		if (buf[0] == '#' || buf[0] == ':')
			continue;

		if (token = strtok(buf, s_delim), token == NULL)
			return false;

		if (token[0] >= '0' && token[0] <= '9')
		{
			frm = static_cast<size_t>(atoi(token)) - 1;
			continue;
		}

		if (jid = fig->FindNode(token), jid < 0)
			return false;

		cit = orders.find(token);
		if (jid == 0)
		{
			parse_root_coord(cit->second, v, q);
			mot->SetPosition(frm, v);
			mot->SetRotation(frm, jid, q);
		}
		else
		{
			parse_joint_rotation(cit->second, q);
			D3DXMatrixRotationQuaternion(&m, &q);
			m = m * calc_axis_transform( fig->GetJoint(jid), axes);
			D3DXQuaternionRotationMatrix(&q, &m);
			mot->SetRotation(frm, jid, q);
		}
	}

	return true;
}

/**
 * @fn void change_skeletal_structure(CJoint *node, D3DXVECTOR3 &v, std::map<std::string, D3DXMATRIX> &axes)
 * @brief ASF`̃XPg\BVH`ɕϊ
 * @param node [in] ֐߃m[hւ̃|C^
 * @param v [in] ItZbgxNg
 * @param axes [in/out] ֐ߖ-֐axiss̑Ή}bv
 */
void change_skeletal_structure(CJoint *node, D3DXVECTOR3 &v, std::map<std::string, D3DXMATRIX> &axes)
{
	static int ls_extent = 0;
	if (node->NumChildren() > 0)
	{
		for (size_t i = 0;i < node->NumChildren();++i)
			change_skeletal_structure(node->GetChild(i), node->GetOffset(), axes);
	}
	else
	{
		CJoint *ext = new CJoint(node->GetOffset(), node);
		node->AddChild(ext);

		char buf[6];
		sprintf(buf, "ext%02d", ls_extent++);
		ext->SetName(buf);

		D3DXMATRIX m;
		D3DXMatrixIdentity(&m);
		axes[buf] = m;
	}
	node->SetOffset(v);
}

/**
 * @fn bool NMotionFile::LoadAsfAmcFile(const char *asfpath, const char *amcpath, CFigure *fig, CMotionData *mot)
 * @brief ASF-AMCt@C̃[h
 * @param asfpath [in] ASFt@Cւ̃pX
 * @param amcpath [in] AMCt@Cւ̃pX
 * @param fig [out] XPgf[^
 * @param mot [out] [Vf[^
 * @retval true [h
 * @retval false [hs
 */
bool NMotionFile::LoadAsfAmcFile(const char *asfpath, const char *amcpath, CFigure *fig, CMotionData *mot)
{
	FILE *fasf, *famc;
	std::map<std::string, int> orders;
	std::map<std::string, D3DXMATRIX> axes;
	CJoint *root;

	if (fasf = fopen(asfpath, "r"), fasf == NULL)
		return false;
	if (!NAsfAmcFile::parse_asf_header(fasf))
		return false;

	if (root = NAsfAmcFile::parse_asf(fasf, orders, axes), root == NULL)
		return false;
	fclose(fasf);

	change_skeletal_structure(root, D3DXVECTOR3(0, 0, 0), axes);
	fig->Initialize(root);

	if (famc = fopen(amcpath, "r"), famc == NULL)
		return false;
	if (!NAsfAmcFile::parse_amc(famc, mot, fig, orders, axes))
		return false;
	fclose(famc);

	return true;
}
