float4x4 g_mWorld : World;
float4x4 g_mWorldIT : WORLDINVERSETRANSPOSE;
float4 g_vEyePos : EyePos;
float4x4 g_mWorldViewProj : WorldViewProj;

const float4 g_vLightPos = float4(-10.0f, 20.0f, 0.0f, 1.0f);
const float4 g_vLightColor = float4(0.8f, 0.8f, 0.6f, 1.0f);
const float g_fSpecExpon = 0.7f;

float4 g_vAmbiColor =float4(0.8f, 0.8f, 0.8f, 1.0f);
float4 g_vSurfColor =float4(0.8f, 0.8f, 0.8f, 1.0f);

struct PHONG_VS_INPUT
{
  float4 vPos : POSITION;
  float3 vNormal : NORMAL;
  float2 vUV : TEXCOORD0;
};

struct PHONG_VS_OUTPUT
{
  float4 vPos : POSITION;
  float3 Normal : TEXCOORD0;
  float3 ViewVec : TEXCOORD1;
  float3 LightVec : TEXCOORD2;
};

PHONG_VS_OUTPUT PhongVS(PHONG_VS_INPUT In)
{
  PHONG_VS_OUTPUT Out = (PHONG_VS_OUTPUT)0;

  float3 Pw = mul(In.vPos, g_mWorld);

  Out.vPos = mul(In.vPos, g_mWorldViewProj);
  Out.ViewVec = normalize(g_vEyePos - Pw);
  Out.LightVec = normalize(g_vLightPos - Pw);
  Out.Normal = mul(In.vNormal, g_mWorldIT);
  Out.Normal = normalize(Out.Normal);

  return Out;
}

float4 PhongPS(PHONG_VS_OUTPUT In) : COLOR0
{
  float3 Hn = normalize(In.ViewVec + In.LightVec);

  float4 litV = lit(dot(In.Normal, In.LightVec), dot(In.Normal, Hn), g_fSpecExpon);
  float4 diffContrib = g_vSurfColor * (0.5 * litV.y * g_vLightColor + g_vAmbiColor);
  float4 specContrib = litV.y * litV.z * g_vLightColor;
  return diffContrib + specContrib;
}

technique PhongShader
{
  pass P0
  {
    VertexShader = compile vs_2_0 PhongVS();
    PixelShader = compile ps_2_0 PhongPS();
  }
}
